from typing import TYPE_CHECKING, Any, Callable, List, Optional

from envium.exceptions import EnviumError

if TYPE_CHECKING:
    pass

from envium.vars import ComputedMixin, Var, VarGroup, VarType

__all__ = ["ctx_var", "Ctx", "computed_ctx_var", "CtxGroup"]


class CtxVar(Var):
    pass


class ComputedCtxVar(ComputedMixin, CtxVar):
    def __init__(
        self, fget: Optional[Callable] = None, fset: Optional[Callable] = None
    ) -> None:
        ComputedMixin.__init__(self, fget=fget, fset=fset)
        CtxVar.__init__(self)

    pass


class CtxGroup(VarGroup[CtxVar]):
    pass


class Ctx(CtxGroup):
    def __init__(self, name: str = ""):
        super().__init__(name=name)
        self._root = self
        self._process()
        a = 1

    def validate(self) -> None:
        self._validate()

    @property
    def errors(self) -> List[EnviumError]:
        return self._errors


def ctx_var(
    default: Optional[Any] = None,
    default_factory: Optional[Callable] = None,
) -> Any:
    return CtxVar(default=default, default_factory=default_factory)


def computed_ctx_var(
    fget: Optional[Callable] = None,
    fset: Optional[Callable] = None,
) -> Any:
    return ComputedCtxVar(fget, fset)


Group = VarGroup



import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

from envium import comp
from envium.exceptions import EnviumError, RedefinedVarError

if TYPE_CHECKING:
    pass

from envium.vars import ComputedMixin, FinalVar, Var, VarGroup, VarType

__all__ = ["env_var", "Environ", "computed_env_var", "EnvGroup"]


class EnvVar(Var):
    raw: Union[bool, str]
    _parent: "EnvGroup"

    def __init__(
        self,
        default: Optional[Any] = None,
        *,
        default_factory: Optional[Callable] = None,
        raw: Union[bool, str] = False,
    ) -> None:
        super().__init__(default=default, default_factory=default_factory)
        self._raw = raw

    def _init_value(self) -> None:
        if self._parent._load:
            env_value = os.environ.get(self._get_env_name(), None)
            env_value = None if env_value == "None" else env_value
            self._value = self._from_str(env_value) if env_value else None

        if self._value is None:
            self._value = self._default

    def _get_env_name(self) -> str:
        if self._raw:
            if self._raw is True:
                ret = self._name
            else:
                ret = self._raw
        else:
            ret = self._fullname
            ret = ret.replace("_", "").replace(".", "_").replace("-", "")

        ret = ret.upper()
        return ret


class ComputedEnvVar(ComputedMixin, EnvVar):
    def __init__(
        self,
        fget: Optional[Callable] = None,
        fset: Optional[Callable] = None,
        *,
        raw: Union[bool, str] = None,
    ) -> None:
        ComputedMixin.__init__(self, fget=fget, fset=fset)
        EnvVar.__init__(self, raw=raw)
    pass


class EnvGroup(VarGroup[EnvVar]):
    raw: Union[bool, str]
    _load: bool

    def __init__(
        self, name: Optional[str] = None, *, raw: Union[bool, str] = False, load: bool = True
    ) -> None:
        super().__init__(name=name)
        self._load = load
        self._raw = raw

    @property
    def _fullname(self) -> str:
        if self._raw:
            if self._raw is True:
                return self._name
            else:
                return self._raw

        ret = super()._fullname
        return ret

    def _dump(self, path: Union[Path, str]) -> None:
        path = Path(path)

        if not path.parent.exists():
            path.parent.mkdir(parents=True)

        path.touch(exist_ok=True)

        content = "\n".join(
            [f'{key}="{value}"' for key, value in self._root._get_env_vars().items()]
        )
        path.write_text(content, "utf-8")


class Environ(EnvGroup):
    def __init__(self, name: str, raw: Union[bool, str] = False, load: bool = False):
        if not name:
            raise EnviumError("Root needs to have a name")

        super().__init__(name=name, load=load, raw=raw)
        self._root = self
        self._process()

        if self._load:
            self._validate()

    def get_env_vars(self) -> Dict[str, str]:
        return self._get_env_vars()

    def validate(self) -> None:
        return self._validate()

    def dump(self, path: Union[Path, str]) -> None:
        return self._dump(path)

    def save_to_os_environ(self) -> None:
        os.environ.update(self.get_env_vars())

    def _get_env_vars(self) -> Dict[str, str]:
        """
        Return environmental variables in following format:
        {NAMESPACE_ENVNAME}

        :param owner_name:
        """

        self._validate()

        envs = {}
        for v in self._flat:
            name = v._get_env_name()
            value = v._get_value()
            if isinstance(value, list):
                envs[name] = comp.list_delimiter.join([str(v) for v in value])
            else:
                envs[name] = str(value)

        envs = {k.upper(): v for k, v in envs.items()}

        return envs

    @property
    def errors(self) -> List[EnviumError]:
        ret: List[EnviumError] = []
        env_names = []
        for v in self._flat:
            env_name = v._get_env_name()

            if env_name in env_names:
                ret.append(RedefinedVarError(env_name))

            env_names.append(env_name)

            ret.extend(v._get_errors())

        return ret


def env_var(
    default: Optional[Any] = None,
    raw: Union[bool, str] = False,
    default_factory: Optional[Callable] = None
) -> Any:
    return EnvVar(default=default, raw=raw, default_factory=default_factory)


def computed_env_var(
    fget: Optional[Callable] = None,
    fset: Optional[Callable] = None,
    raw: Union[bool, str] = False
) -> Any:
    return ComputedEnvVar(fget, fset, raw=raw)


Group = VarGroup



import sys

PLATFORM_WINDOWS = "windows"
PLATFORM_LINUX = "linux"
PLATFORM_BSD = "bsd"
PLATFORM_DARWIN = "darwin"
PLATFORM_UNKNOWN = "unknown"


def get_platform_name() -> str:
    if sys.platform.startswith("win"):
        return PLATFORM_WINDOWS
    elif sys.platform.startswith("darwin"):
        return PLATFORM_DARWIN
    elif sys.platform.startswith("linux"):
        return PLATFORM_LINUX
    elif sys.platform.startswith(("dragonfly", "freebsd", "netbsd", "openbsd", "bsd")):
        return PLATFORM_BSD
    else:
        return PLATFORM_UNKNOWN


__platform__ = get_platform_name()


def is_linux() -> bool:
    return __platform__ == PLATFORM_LINUX


def is_bsd() -> bool:
    return __platform__ == PLATFORM_BSD


def is_darwin() -> bool:
    return __platform__ == PLATFORM_DARWIN


def is_windows() -> bool:
    return __platform__ == PLATFORM_WINDOWS


if is_windows():
    list_delimiter = ";"
else:
    list_delimiter = ":"



from typing import List, Type


class EnviumError(Exception):
    pass


class RedefinedVarError(EnviumError):
    def __init__(self, var_name: str) -> None:
        self.var_name = var_name
        msg = f'Variable "{var_name}" is redefined'
        super().__init__(msg)


class WrongTypeError(EnviumError):
    def __init__(self, var_name: str, type_: Type, got_type: Type) -> None:
        self.var_name = var_name
        self.type_ = type_
        self.got_type = got_type

        msg = (
            f'Expected type "{repr(type_)}" for var "{var_name}" got "{repr(got_type)}"'
        )
        super().__init__(msg)


class NoTypeError(EnviumError):
    def __init__(self, var_name: str) -> None:
        self.var_name = var_name
        msg = f'Type annotation for var "{var_name}" is missing'
        super().__init__(msg)


class NoValueError(EnviumError):
    def __init__(self, var_name: str, type_: Type) -> None:
        self.var_name = var_name
        self.type_ = type_
        msg = f'Expected value of type "{repr(type_)}" for var "{var_name}" not None'
        super().__init__(msg)


class ComputedVarError(EnviumError):
    def __init__(self, var_name: str, exception: Exception) -> None:
        self.var_name = var_name
        msg = f'During computing "{var_name}" following error occured: \n{repr(exception)}'
        super().__init__(msg)


class UndefinedVarError(EnviumError):
    def __init__(self, parent_fullname: str, var_name: str) -> None:
        self.var_name = var_name
        msg = f'Var grup "{parent_fullname}" does not have var "{var_name}"'
        super().__init__(msg)


class ValidationErrors(EnviumError):
    errors: List[EnviumError]

    def __init__(self, errors: List[EnviumError]) -> None:
        self.errors = errors
        msg = "\n".join(repr(e) for e in self.errors)
        super().__init__(msg)



import inspect
import os

# Python >= 3.8
import typing
from abc import ABC, abstractmethod
from copy import deepcopy
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from envium import comp
from envium.exceptions import (
    ComputedVarError,
    EnviumError,
    NoTypeError,
    NoValueError,
    RedefinedVarError,
    UndefinedVarError,
    ValidationErrors,
    WrongTypeError,
)

try:
    typing.get_args  # type: ignore
    typing.get_origin  # type: ignore
# Compatibility
except AttributeError:
    typing.get_args = lambda t: getattr(t, "__args__", ()) if t is not Generic else Generic  # type: ignore
    typing.get_origin = lambda t: getattr(t, "__origin__", None)  # type: ignore


__all__ = ["VarGroup"]

VarType = TypeVar("VarType", bound="FinalVar")


def get_type_class(typ) -> Any:
    try:
        # Python 3.5 / 3.6
        return typ.__extra__
    except AttributeError:
        # Python 3.7
        if hasattr(typ, "__origin__"):
            return typ.__origin__
        else:
            return typ


class BaseVar(ABC, Generic[VarType]):
    # will be injected by parent
    _root: Optional["VarGroup"]
    _parent: Optional["BaseVar"]
    _name: str
    _ready: bool

    def __init__(self) -> None:
        self._root = None
        self._parent = None
        self._name = ""
        self._ready = False

    @property
    def _fullname(self) -> str:
        ret = f"{self._parent._fullname}.{self._name}" if self._parent else self._name
        return ret


class FinalVar(BaseVar, ABC, Generic[VarType]):
    _type_: Optional[Type]
    _optional: bool
    _value: Optional[VarType]

    _final: ClassVar[bool] = True
    _ready: bool

    def __init__(self) -> None:
        super().__init__()

        self._type_ = None
        self._optional = False
        self._value = None

    @abstractmethod
    def _init_value(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def _get_errors(self) -> List[EnviumError]:
        return []

    def __repr__(self) -> str:
        return f"{self._fullname}"

    @abstractmethod
    def _get_value(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def _set_value(self, new_value) -> None:
        raise NotImplementedError


class Var(FinalVar, Generic[VarType]):
    _default: Optional[VarType]
    _default_factory: Optional[Callable]

    def __init__(
        self,
        default: Optional[VarType] = None,
        default_factory: Optional[Callable] = None,
    ) -> None:
        super().__init__()
        self._default_factory = default_factory
        self._default = default

        if self._default_factory:
            self._default = self._default_factory()

    def _init_value(self):
        if not self._value:
            self._value = self._default

    def _get_value(self) -> Any:
        return self._value

    def _set_value(self, new_value) -> None:
        self._value = new_value

    def _get_errors(self) -> List[EnviumError]:
        ret: List[EnviumError] = []

        # Try evaluating value first. There might be some issues with that
        if not self._type_:
            return [NoTypeError(var_name=self._fullname)]

        if not self._optional and self._get_value() is None:
            ret.append(NoValueError(type_=self._type_, var_name=self._fullname))
        elif self._get_value() is not None:
            try:
                if self._type_ and not isinstance(self._get_value(), self._type_):
                    ret.append(
                        WrongTypeError(
                            type_=self._type_,
                            var_name=self._fullname,
                            got_type=type(self._get_value()),
                        )
                    )
            except TypeError:
                # isinstance will fail for types like Union[] etc
                pass

        return super()._get_errors() + ret

    def _from_str(self, env_value: str) -> VarType:
        ret: Any
        if self._type_ is bool:
            ret = env_value in ("True", "true")
        elif issubclass(get_type_class(self._type_), list):
            ret = env_value.split(comp.list_delimiter)
        else:
            ret = self._type_(env_value)

        ret_casted = cast(VarType, ret)
        return ret_casted


class ComputedMixin(Var):
    _fget: Optional[Callable]
    _fset: Optional[Callable]

    def __init__(
        self,
        fget: Optional[Callable] = None,
        fset: Optional[Callable] = None,
    ) -> None:
        super().__init__()
        self._fget = fget
        self._fset = fset

    def _init_value(self):
        self._value = self._get_value()

    def _get_value(self) -> Any:
        object.__setattr__(self, "_ready", False)
        if self._fget:
            ret = self._fget(self._root)
        else:
            ret = self._value
        object.__setattr__(self, "_ready", True)
        return ret

    def _set_value(self, new_value) -> None:
        object.__setattr__(self, "_ready", False)
        if self._fset:
            self._fset(self._root, new_value)
        else:
            self._value = new_value
        object.__setattr__(self, "_ready", True)

    def _get_errors(self) -> List[EnviumError]:
        try:
            self._get_value()
        except Exception as e:
            return [ComputedVarError(var_name=self._fullname, exception=e)]

        return super()._get_errors()


class VarGroup(BaseVar[VarType]):
    _children: List[VarType]

    def __init__(self, name: str = ""):
        super().__init__()
        self._children = []
        self._name = name

        # Create copy of the var class attributes and assign them to the instance
        for f in dir(self):
            attr = inspect.getattr_static(self, f)

            if isinstance(attr, BaseVar):
                setattr(self, f, deepcopy(attr))

    def _process(self) -> None:
        self._children.clear()

        annotations = [
            c.__annotations__
            for c in self.__class__.__mro__
            if hasattr(c, "__annotations__")
        ]
        flat_annotations = {}

        for a in reversed(annotations):
            flat_annotations.update(a)

        for n in dir(self):
            v = inspect.getattr_static(self, n)

            if v is self._root:
                continue

            if v is self._parent:
                continue

            if not isinstance(v, BaseVar):
                continue

            v = cast(TypeVar, v)

            v._name = n
            v._root = self._root
            v._parent = self

            self._children.append(v)

            if isinstance(v, VarGroup):
                v._process()
            elif isinstance(v, FinalVar):
                type_ = flat_annotations.get(n, None)
                optional = typing.get_origin(type_) is Union and type(None) in typing.get_args(type_)  # type: ignore
                v._optional = optional

                if optional:
                    v._type_ = typing.get_args(type_)[0]  # type: ignore
                else:
                    v._type_ = type_
                v._init_value()

            v._ready = True

        self._ready = True

    @property
    def _flat(self) -> List[VarType]:
        ret: List[VarType] = []
        for c in self._children:
            if isinstance(c, VarGroup):
                ret.extend(c._flat)
            elif isinstance(c, FinalVar):
                c = cast(VarType, c)
                ret.append(c)

        ret = sorted(ret, key=lambda x: x._fullname)
        return ret

    def copy_from(self, var_group: "VarGroup") -> None:
        left = {v._name: v for v in self._children}
        right = {v._name: v for v in var_group._children}

        for k, l in left.items():
            r = right.get(k)

            if l.__class__ is not r.__class__:
                continue

            if isinstance(l, VarGroup):
                l.copy_from(r)
            else:
                l._value = r._value

    def __setattr__(self, key: str, value: Any) -> None:
        if not hasattr(self, key):
            if hasattr(self, "_ready") and self._ready:
                raise UndefinedVarError(parent_fullname=self._fullname, var_name=key)
            object.__setattr__(self, key, value)
            return

        attr = object.__getattribute__(self, key)

        if not isinstance(attr, FinalVar):
            object.__setattr__(self, key, value)
            return

        if not attr._ready:
            object.__setattr__(self, key, value)
            return
        else:
            attr._set_value(value)

    def __getattribute__(self, item: str) -> Any:
        attr = object.__getattribute__(self, item)

        if not isinstance(attr, FinalVar):
            return attr

        if not attr._ready:
            return attr

        return attr._get_value()

    @property
    def _errors(self) -> List[EnviumError]:
        ret: List[EnviumError] = []
        for v in self._flat:
            ret.extend(v._get_errors())

        return ret

    def _validate(self) -> None:
        errors = self._errors

        if errors:
            raise ValidationErrors(errors)
